//
// anyRemote
// a bluetooth remote for your PC.
//
// Copyright (C) 2012-2012 Mikhail Fedotov <anyremote@mail.ru>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "common.h"
#include "hash.h"
#include "list.h"

//
// Hash related wrappers
//

#ifdef USE_GLIB

#include <glib.h>

GHashTable* _hashes[H_MAX] = {NULL};

int hashNew (int id, DestroyCallback valueDestroyFunc)
{
     if (id >=0 && id < H_MAX && _hashes[id] == NULL) {
         _hashes[id] = g_hash_table_new_full(g_str_hash, g_str_equal, free, valueDestroyFunc);
	 return RC_OK;
     }
     return RC_NOK;
}

void* hashFind(int id, const char *key)
{
    if (id >=0 && id < H_MAX && _hashes[id] != NULL) {
        return g_hash_table_lookup(_hashes[id], key);
    }
    return NULL;
}

void hashReplace(int id, const char* key, void *value)
{
    if (id >=0 && id < H_MAX && _hashes[id] != NULL) {
        g_hash_table_replace(_hashes[id], strdup(key), value);
    }
}

int hashRemove(int id, const char* key)
{
    if (id >=0 && id < H_MAX && _hashes[id] != NULL) {
        if (g_hash_table_remove(_hashes[id], key)) {
	        return RC_OK;
	    }
    }
    return RC_NOK;
}

void hashDestroy(int id)
{
    if (id >=0 && id < H_MAX && _hashes[id] != NULL) {
        g_hash_table_destroy(_hashes[id]);
	    _hashes[id] = NULL;
    }
}

void hashForeach(int id, HashForeachFunc func, void *data)
{
    if (id >=0 && id < H_MAX && _hashes[id] != NULL) {
        g_hash_table_foreach(_hashes[id], func, data);
    }
}

#else

#define NR_BUCKETS 1024

struct StrHashNode {
    char *key;
    void *value;
    struct StrHashNode *next;
};

struct StrHashTable {

    struct StrHashNode* buckets[NR_BUCKETS];
    DestroyCallback     free_value;
};

static unsigned int str2hash(const char *str)
{
    unsigned int hash = 0;
    for(; *str; str++) {
        hash = 31*hash + *str;
    }
    return hash;
}

static void* get(struct StrHashTable *table, const char *key)
{
    unsigned int bucket = str2hash(key)%NR_BUCKETS;
    struct StrHashNode *node = table->buckets[bucket];
    
    while (node) {
        if (strcmp(key,node->key) == 0) {
            return node->value;
        }
	node = node->next;
    }
    return NULL;
}

static int insert(struct StrHashTable *table, char *key, void *value)
{
    
    unsigned int bucket = str2hash(key)%NR_BUCKETS;
    //printf("insert %s to %d\n", key, bucket);

    struct StrHashNode **tmp = &(table->buckets[bucket]);
    while (*tmp) {
        if (strcmp(key,(*tmp)->key) == 0) {
            break;
        }
	tmp = &((*tmp)->next);
    }
    
    struct StrHashNode *node = NULL;
    
    if (*tmp) {
        //printf("insert %s remove old\n", key);
        free((*tmp)->key);
 
	if(table->free_value != NULL) {
            table->free_value((*tmp)->value);
        }
	node = *tmp;
    } else {
        node = malloc(sizeof(struct StrHashNode));
        if(node == NULL) {
            return -1;
        }
	node->next = NULL;
        *tmp = node;
    }
    node->key   = key;
    node->value = value;

    return 0;
}

static int removeNode(struct StrHashTable *table, const char *key)
{
    unsigned int bucket = str2hash(key)%NR_BUCKETS;
    
    struct StrHashNode *tmp;
    struct StrHashNode *tmp_prev;
    
    tmp_prev = NULL;
    tmp = table->buckets[bucket];
    while (tmp) {
        if (strcmp(key,tmp->key) == 0) {
            break;
        }
	tmp_prev = tmp;
	tmp = tmp->next;
    }
    
    if(tmp) {
    
        free(tmp->key);

	if(table->free_value != NULL) {
            table->free_value(tmp->value);
        }
	
	if (tmp_prev) {
	    tmp_prev->next = tmp->next;
	} else {  // remove first
	    table->buckets[bucket] = tmp->next;
	}
	
	free(tmp);
	
	return RC_OK;
    }

    return RC_NOK;
}

static void destroyData(struct StrHashTable *table)
{
    int i = 0;
    for (;i<NR_BUCKETS;i++) {

	struct StrHashNode *tmp = table->buckets[i];
	
	while (tmp) {
	    
            free(tmp->key);

	    if(table->free_value != NULL) {
                table->free_value(tmp->value);
            }
            
	    struct StrHashNode *t = tmp;
	   
	    tmp = tmp->next;
	    
	    free(t);
	}
    }
}

static void foreach(struct StrHashTable *table, HashForeachFunc func, void *data)
{
    int i = 0;
    for (;i<NR_BUCKETS;i++) {

	struct StrHashNode *tmp = table->buckets[i];   
	if (!tmp) continue;
	
	//printf("foreach %d\n", i);
	
	while (tmp) {
	    if (tmp->key && tmp->value) {
	        func(tmp->key, tmp->value, data);
	    //} else {
	    //    printf("foreach %d corrupted hash data\n", i);
	    }
	    tmp = tmp->next;
	}
    }
}

struct StrHashTable* _hashes[H_MAX] = {NULL};

int hashNew(int id, DestroyCallback valueDestroyFunc)
{
     if (id >=0 && id < H_MAX && _hashes[id] == NULL) {
         _hashes[id] = malloc(sizeof(struct StrHashTable));
    
         int i = 0;
	 for (;i<NR_BUCKETS;i++) {
	     _hashes[id]->buckets[i]  = NULL;
	 }
	 _hashes[id]->free_value = valueDestroyFunc;

	 return RC_OK;
     }
     return RC_NOK;
}

void* hashFind(int id, const char *key)
{
    if (id >=0 && id < H_MAX && _hashes[id] != NULL) {
        return get(_hashes[id], key);
    }
    return NULL;
}

void hashReplace(int id, const char* key, void *value)
{
    if (id >=0 && id < H_MAX && _hashes[id] != NULL) {
        insert(_hashes[id], strdup(key), value);
    }
}

int hashRemove(int id, const char* key)
{
    if (id >=0 && id < H_MAX && _hashes[id] != NULL) {
        return removeNode(_hashes[id], key);
    }
    return RC_NOK;
}

void hashDestroy(int id)
{
    if (id >=0 && id < H_MAX && _hashes[id] != NULL) {
        destroyData(_hashes[id]);
	free(_hashes[id]);
	_hashes[id] = NULL;
    }
}

void hashForeach(int id, HashForeachFunc func, void *data)
{
    if (id >=0 && id < H_MAX && _hashes[id] != NULL) {
        foreach(_hashes[id], func, data);
    }
}

#endif

int hashExists(int id)
{
    return (id >=0 && id < H_MAX && _hashes[id] != NULL ? RC_OK : RC_NOK);
}
